import os
import sys

from . import versions

CURRDIR = versions.get_repo_dir()
assert CURRDIR, "Not a valid Git repository"
PROJID = os.path.basename(CURRDIR)
HOMEDIR = os.path.join(os.path.expanduser("~"), ".flor")
os.makedirs(HOMEDIR, exist_ok=True)
SCRIPTNAME = os.path.basename(sys.argv[0])

__all__ = [
    "CURRDIR",
    "PROJID",
    "HOMEDIR",
    "SCRIPTNAME",
]



import time
from datetime import datetime
from . import cli


class Clock:
    start_time = time.perf_counter()
    current_datetime = datetime.now().isoformat(timespec="seconds")

    def __init__(self) -> None:
        self.s_time = None

    def set_start_time(self):
        self.s_time = time.perf_counter()

    def get_delta(self):
        s_time = self.start_time if self.s_time is None else self.s_time
        return time.perf_counter() - s_time

    @classmethod
    def get_datetime(cls):
        if cli.in_replay_mode():
            assert cli.flags.old_tstamp is not None
            return cli.flags.old_tstamp
        else:
            assert cls.current_datetime is not None
            return cls.current_datetime

    @classmethod
    def set_new_datetime(cls):
        cls.current_datetime = datetime.now().isoformat(timespec="seconds")
        cls.start_time = time.perf_counter()



# type: ignore
from collections.abc import MutableMapping
from typing import Generic, TypeVar, Iterable, NamedTuple


K, V = TypeVar("K"), TypeVar("V")


class IdMap(MutableMapping):
    def __init__(self, it: Iterable["tuple[K, V]"] = ()):
        self.Item = NamedTuple("Item", [("key", K), ("value", V)])
        self._map = {}
        for k, v in it:
            self[k] = v

    def __getitem__(self, key: K) -> V:
        return self._map[id(key)].value

    def __setitem__(self, key: K, value: V):
        self._map[id(key)] = self.Item(key, value)  # type: ignore

    def __delitem__(self, key: K):
        del self._map[id(key)]

    def __iter__(self) -> Iterable[K]:
        for item in self._map.values():
            yield item.key

    def __len__(self) -> int:
        return len(self._map)

    def items(self) -> Iterable["tuple[K, V]"]:
        return self._map.values()



from dataclasses import dataclass, field
from .adapter import BaseAdapter


@dataclass(frozen=True)
class Node:
    label: str
    value: str = ''
    children: list['Node'] = field(default_factory=list)
    parent: 'Node' = field(init=False, compare=False, repr=False)

    def __post_init__(self):
        for child in self.children:
            object.__setattr__(child, 'parent', self)
        super().__setattr__('parent', None)
    
    def __getitem__(self, i):  # testing
        return self.children[i]

Tree = Node


class Adapter(BaseAdapter[Tree]):
    def parent(self, n: Tree):
        return n.parent

    def children(self, n: Tree):
        return n.children

    def label(self, n: Tree) -> str:
        return n.label

    def value(self, n: Tree) -> str:
        return n.value
    
    def isomorphic(self, n1: Tree, n2: Tree) -> bool:
        return n1 == n2


adapter = Adapter()



#!/usr/bin/env python

from argparse import Namespace
from ast import AST, Name, parse, unparse, walk, stmt, literal_eval
from collections import defaultdict
from copy import deepcopy
import ast

from .gumtree import GumTree, python


def propagate(args: Namespace):
    tree, target = [parse(f.read()) for f in (args.source, args.target)]
    replicate(tree, find(tree, lineno=args.lineno), target, **args.gumtree)
    with open(args.out, "w") as f:
        print(unparse(target), file=f)


def replicate(tree: AST, node: AST, target: AST, **kwargs):
    adapter = python.Adapter(tree, target)
    mapping = GumTree(adapter, **kwargs).mapping(tree, target)
    assert tree == adapter.root(node) and isinstance(node, stmt)

    if node in mapping:
        raise FileExistsError("Nothing to do")

    block, index = find_insert_loc(adapter, node, mapping)
    new = make_contextual_copy(adapter, node, mapping)
    block.insert(index, new)  # type: ignore


def find_insert_loc(adapter, node, mapping):
    parent = adapter.parent(node)

    # Find the context by checking the mapping of sibling nodes
    context = None
    for sibling in adapter.children(parent):
        if id(sibling) == id(node):
            break
        if sibling in mapping:
            context = sibling

    # Determine the block and index for insertion based on the context found
    if context is not None:
        ref = mapping[context]
        block = adapter.parent(ref)
        index = 1 + block.index(ref)
    elif parent in mapping:
        block = mapping[parent]
        index = 0
    else:
        raise ValueError("Unable to map context!")  # Raise a specific exception

    assert block is not None
    return block, index


def make_contextual_copy(adapter, node, mapping):
    renames = defaultdict(lambda: defaultdict(int))
    for source, target in mapping.items():
        if isinstance(source, ast.Name) and not adapter.contains(source, node):
            renames[source.id][target.id] += 1

    new_node = deepcopy(node)

    for n in ast.walk(new_node):
        if isinstance(n, ast.Name) and n.id in renames:
            most_common_target_id = max(
                renames[n.id].items(), key=lambda item: item[1]
            )[0]
            n.id = most_common_target_id

    return new_node


class FindNodeVisitor(ast.NodeVisitor):
    def __init__(self, lineno):
        self.lineno = lineno
        self.result = None

    def generic_visit(self, node):
        if getattr(node, "lineno", None) == self.lineno:
            self.result = node
        else:
            super().generic_visit(node)


def find(t: AST, lineno: int):
    visitor = FindNodeVisitor(lineno)
    visitor.visit(t)
    assert visitor.result is not None

    return visitor.result



from .constants import *
from .cli import flags
from . import database
from . import versions
from . import repl


import json
import platform
import sys
import os
import atexit


def main():
    if flags.args is not None and flags.args.flor_command is not None:
        if flags.args.flor_command == "unpack":
            conn, cursor = database.conn_and_cursor()
            database.create_tables(cursor)

            start_branch = versions.current_branch()
            assert start_branch is not None
            known_tstamps = [t for t, in database.read_known_tstamps(cursor)]
            try:
                for triplet in versions.get_latest_autocommit():
                    ts_start, next_commit, _ = triplet
                    if ts_start in known_tstamps:
                        break
                    versions.checkout(next_commit)
                    with open(".flor.json", "r") as f:
                        database.unpack(json.load(f), cursor)
                conn.commit()
                conn.close()
            finally:
                versions.checkout(start_branch.name)
        elif flags.args.flor_command == "query":
            user_query = str(flags.args.q)
            df = repl.query(user_query)
            print(df)
        elif flags.args.flor_command == "dataframe":
            df = repl.dataframe(
                *(flags.args.columns if flags.args.columns else tuple())
            )
            print(df)
        elif flags.args.flor_command == "replay":
            if flags.args.where_clause:
                repl.replay(flags.args.VARS, flags.args.where_clause)
            else:
                repl.replay(flags.args.VARS)
        elif flags.args.flor_command == "stat":
            build_context = {
                "architecture": platform.machine(),
                "operating_system": platform.system(),
                "os_version": platform.version(),
                "python_version": platform.python_version(),
                "in_anaconda": "conda" in sys.version
                or "Continuum" in sys.version
                or os.path.exists(os.path.join(sys.prefix, "conda-meta")),
            }
            for k, v in build_context.items():
                print(k + ":", v)


if __name__ == "__main__":
    main()
